/*


    /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
   /                                                                      \
   \      ___   _____   _____   _   __           _____   _____  __  __    /
   /     |_  | |___  | |  _  | | | / /          |  _  \ |  _  | \ \/ /    \
   \       | |  ___| | | | |_| | |/ /           | |_/ / | | | |  \  /     /
   /    _  | | |  _  | | |  _  |   <    in the  |  _ <  | | | |   ><      \
   \   | |_| | | |_| | | |_| | | |\ \           | |_| \ | |_| |  /  \     /
   /   |_____| |_____| |_____| |_| \_\          |_____/ |_____| /_/\_\    \
   \                                                                      /
   /                                                                      \
   \                                                                      /
   /             C    O    M    P    U    T    I    N    G                \
   \                                                                      /
    \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

*/


#define EXTERN extern
#include "game.h"

//============================================================================
// Module  : JOYSTICK
// Contents: joystick controlling
// Date    : 06.09.1997
// Updates : (none)
//============================================================================

//@CUSTOMIZE@ max number if joystick timeouts before disconnecting it
#define JOYSTICK_MAX_TIMEOUTS 3

//@CUSTOMIZE@ frenquency in hz to read the joystick buttons
#define READ_BUTTONS_HZ 10

//@CUSTOMIZE@ frenquency in hz to read the joystick XY axis
#define READ_AXIS_HZ 4

//@CUSTOMIZE@ frequency in milliseconds to look for joystick re-connections
#define CONNECTION_PULL_TIME_MS 2000

//@CUSTOMIZE@ tolerance from centered joystick position to determine 
//            movements (this default means 30%)
#define AXIS_TOLERANCE 0.3

//tolerance aux formulas
#define AXIS_TOLERANCE_POSITIVE (1.0 + AXIS_TOLERANCE)
#define AXIS_TOLERANCE_NEGATIVE (1.0 - AXIS_TOLERANCE)

//joystick 1 connected flag
volatile int nJoystick1Connected = FALSE;

//joystick 1 timeout counter
volatile int nJoystick1Timeouts = 0;

//joystick 1 buttons
volatile int nJoystick1A, nJoystick1B;

//joystick 1 axis
volatile int nJoystick1X, nJoystick1Y;

//joystick 1 callibration
volatile int nJoystick1CenterX, nJoystick1CenterY;

//joystick 1 direction flags
volatile int nJoystick1Up,   nJoystick1Down;
volatile int nJoystick1Left, nJoystick1Right;

//joystick 1 timed events
volatile int nJoystick1Event1 = -1;
volatile int nJoystick1Event2 = -1;

//joystick 2 connected flag
volatile int nJoystick2Connected = FALSE;

//joystick 2 timeout counter
volatile int nJoystick2Timeouts = 0;

//joystick 2 buttons
volatile int nJoystick2A, nJoystick2B;

//joystick 2 axis
volatile int nJoystick2X, nJoystick2Y;

//joystick 2 callibration
volatile int nJoystick2CenterX, nJoystick2CenterY;

//joystick 2 direction flags
volatile int nJoystick2Up,   nJoystick2Down;
volatile int nJoystick2Left, nJoystick2Right;

//joystick 2 timed events
volatile int nJoystick2Event1 = -1;
volatile int nJoystick2Event2 = -1;

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mReadJoystick1Buttons
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : read joystick 1 buttons A and B
// IN       : (none)
// OUT      : (none)
// Notes    : uses module variables nJoystick1A and nJoystick1B
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mReadJoystick1Buttons() {

   //joystick connected?
   if (nJoystick1Connected == FALSE)
      return;

   //read buttons from joystick 1
   asm volatile ("

         xorl %%eax, %%eax
         movl %%eax, %0
         movl %%eax, %1

         movw $0x201, %%dx
         outb %%al,   %%dx
         inb  %%dx,   %%al
         mov  %%al,   %%ah

      Joystick1ButtonA:

         andb $16, %%al
         jne  Joystick1ButtonB
         movl $1, %0

      Joystick1ButtonB:

         andb $32, %%ah
         jne  Joystick1ButtonExit
         movl $1, %1

      Joystick1ButtonExit:"

      : "=g" (nJoystick1A), "=g" (nJoystick1B)
      : /* nothing */
      : "%eax", "%edx"
   );
}
END_OF_FUNCTION(mReadJoystick1Buttons);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mReadJoystick2Buttons
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : read joystick 2 buttons A and B
// IN       : (none)
// OUT      : (none)
// Notes    : uses module variables nJoystick2A and nJoystick2B
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mReadJoystick2Buttons() {

   //joystick connected?
   if (nJoystick2Connected == FALSE)
      return;

   //read buttons from joystick 2
   asm volatile ("

         xorl %%eax, %%eax
         movl %%eax, %0
         movl %%eax, %1

         movw $0x201, %%dx
         outb %%al,   %%dx
         inb  %%dx,   %%al
         mov  %%al,   %%ah

      Joystick2ButtonA:

         andb $64, %%al
         jne  Joystick2ButtonB
         movl $1, %0

      Joystick2ButtonB:

         andb $128, %%ah
         jne  Joystick2ButtonExit
         movl $1, %1

      Joystick2ButtonExit:"

      : "=g" (nJoystick2A), "=g" (nJoystick2B)
      : /* nothing */
      : "%eax", "%edx"
   );
}
END_OF_FUNCTION(mReadJoystick2Buttons);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mReadJoystick1Axis
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : read joystick 1 X and Y axis
// IN       : (none)
// OUT      : (none)
// Notes    : uses module variables nJoystick1X and nJoystick1Y
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mReadJoystick1Axis() {

   //read axis from joystick 1
   asm volatile ("

         xorl %%eax,  %%eax
         xorl %%ebx,  %%ebx
         movw $0x201, %%dx
         outb %%al,   %%dx

      Joystick1LoopXY:

         cmpl $0x3000, %%ebx
         je   Joystick1TimeOut
         incl %%ebx

         inb  %%dx, %%al
         movb %%al, %%ah
         andb $1,   %%al
         je   Joystick1AxisX

         movb %%ah, %%al
         andb $2,   %%al
         je   Joystick1AxisY

         jmp  Joystick1LoopXY

      Joystick1AxisX:
         movl %%ebx, %0

         Joystick1LoopY:

            cmpl $0x3000, %%ebx
            je   Joystick1TimeOut
            incl %%ebx

            inb  %%dx, %%al
            andb $2,   %%al
            jne  Joystick1LoopY

            movl %%ebx, %1
            jmp  Joystick1Done

      Joystick1AxisY:
         movl %%ebx, %1

         Joystick1LoopX:

            cmpl $0x3000, %%ebx
            je   Joystick1TimeOut
            incl %%ebx

            inb  %%dx, %%al
            andb $1,   %%al
            jne  Joystick1LoopX

            movl %%ebx, %0
            jmp  Joystick1Done

      Joystick1TimeOut:

         movl $0x3000, %%ebx
         movl %%ebx,   %0
                          
      Joystick1Done:"

      : "=g" (nJoystick1X), "=g" (nJoystick1Y)
      : /* nothing */
      : "%eax", "%ebx", "%edx"
   );

   //look for timeouts
   if (nJoystick1X == 0x3000) {

      //reset values
      nJoystick1X = nJoystick1CenterX;
      nJoystick1Y = nJoystick1CenterY;

      //increase timeouts
      nJoystick1Timeouts++;

      //disconnected?
      if (nJoystick1Timeouts >= JOYSTICK_MAX_TIMEOUTS) {

         //disconnect
         nJoystick1Connected = FALSE;

         //change event time for disconnected joysticks (pull)
         mChangeEventTime(nJoystick1Event1, CONNECTION_PULL_TIME_MS);
         mChangeEventTime(nJoystick1Event2, CONNECTION_PULL_TIME_MS);
      }
   }
   else {

      //disconnected?
      if (nJoystick1Connected == FALSE) {

         //connect
         nJoystick1Connected = TRUE;

         //get a new callibration
         nJoystick1CenterX = nJoystick1X;
         nJoystick1CenterY = nJoystick1Y;

         //change event time for connected joysticks
         mChangeEventTime(nJoystick1Event1, HzToMs(READ_BUTTONS_HZ));
         mChangeEventTime(nJoystick1Event2, HzToMs(READ_AXIS_HZ));
      }

      //reset timeouts
      nJoystick1Timeouts = 0;
   }
}
END_OF_FUNCTION(mReadJoystick1Axis);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mReadJoystick2Axis
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : read joystick 2 X and Y axis
// IN       : (none)
// OUT      : (none)
// Notes    : uses module variables nJoystick2X and nJoystick2Y
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mReadJoystick2Axis() {

   //read axis from joystick 2
   asm volatile ("

         xorl %%eax,  %%eax
         xorl %%ebx,  %%ebx
         movw $0x201, %%dx
         outb %%al,   %%dx

      Joystick2LoopXY:

         cmpl $0x3000, %%ebx
         je   Joystick2TimeOut
         incl %%ebx

         inb  %%dx, %%al
         movb %%al, %%ah
         andb $4,   %%al
         je   Joystick2AxisX

         movb %%ah, %%al
         andb $8,   %%al
         je   Joystick2AxisY

         jmp  Joystick2LoopXY

      Joystick2AxisX:
         movl %%ebx, %0

         Joystick2LoopY:

            cmpl $0x3000, %%ebx
            je   Joystick2TimeOut
            incl %%ebx

            inb  %%dx, %%al
            andb $4,   %%al
            jne  Joystick2LoopY

            movl %%ebx, %1
            jmp  Joystick2Done

      Joystick2AxisY:
         movl %%ebx, %1

         Joystick2LoopX:

            cmpl $0x3000, %%ebx
            je   Joystick2TimeOut
            incl %%ebx

            inb  %%dx, %%al
            andb $8,   %%al
            jne  Joystick2LoopX

            movl %%ebx, %0
            jmp  Joystick2Done

      Joystick2TimeOut:

         movl $0x3000, %%ebx
         movl %%ebx,   %0
                          
      Joystick2Done:"

      : "=g" (nJoystick2X), "=g" (nJoystick2Y)
      : /* nothing */
      : "%eax", "%ebx", "%edx"
   );

   //look for timeouts
   if (nJoystick2X == 0x3000) {

      //reset values
      nJoystick2X = nJoystick2CenterX;
      nJoystick2Y = nJoystick2CenterY;

      //increase timeouts
      nJoystick2Timeouts++;

      //disconnected?
      if (nJoystick2Timeouts >= JOYSTICK_MAX_TIMEOUTS) {

         //disconnect
         nJoystick2Connected = FALSE;

         //change event time for disconnected joysticks (pull)
         mChangeEventTime(nJoystick2Event1, CONNECTION_PULL_TIME_MS);
         mChangeEventTime(nJoystick2Event2, CONNECTION_PULL_TIME_MS);
      }
   }
   else {

      //disconnected?
      if (nJoystick2Connected == FALSE) {

         //connect
         nJoystick2Connected = TRUE;

         //get a new callibration
         nJoystick2CenterX = nJoystick2X;
         nJoystick2CenterY = nJoystick2Y;

         //change event time for connected joysticks
         mChangeEventTime(nJoystick2Event1, HzToMs(READ_BUTTONS_HZ));
         mChangeEventTime(nJoystick2Event2, HzToMs(READ_AXIS_HZ));
      }

      //reset timeouts
      nJoystick2Timeouts = 0;
   }
}
END_OF_FUNCTION(mReadJoystick2Axis);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mCallibrateJoystick
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : makes joystick center callibration
// IN       : nJoystick = number of thee joystick (1 or 2)
// OUT      : TRUE or FALSE if joystick is connected
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
int mCallibrateJoystick(int nJoystick) {

   //valid joystick number?
   if ((nJoystick != 1)&&(nJoystick != 2))
      return(FALSE);

   //which one?
   switch(nJoystick) {

      //joystick 1
      case 1:

         //reset values
         nJoystick1Connected = FALSE;
         nJoystick1Timeouts  = JOYSTICK_MAX_TIMEOUTS;

         //read joystick 1 axis
         DISABLE();
         mReadJoystick1Axis();
         ENABLE();

      break;

      //joystick 2
      case 2:

         //reset values
         nJoystick2Connected = FALSE;
         nJoystick2Timeouts  = JOYSTICK_MAX_TIMEOUTS;

         //read joystick 2 axis
         DISABLE();
         mReadJoystick2Axis();
         ENABLE();

      break;
   }

   //success!!!
   return(TRUE);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mInstallJoystick
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : install joystick functions using timer interruptions
// IN       : nJoystick = number of the joystick to install (1 or 2)
// OUT      : TRUE or FALSE if could install joystick
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
int mInstallJoystick(int nJoystick) {

   //wrong number of joysticks?
   if ((nJoystick != 1)&&(nJoystick != 2))
      return(FALSE);

   //which one?
   switch(nJoystick) {
      
      //install timed events for joystick 1:
      case 1:
   
         //lock my data and code
         LOCK_VARIABLE(nJoystick1Connected);
         LOCK_VARIABLE(nJoystick1Timeouts);
         LOCK_VARIABLE(nJoystick1A);
         LOCK_VARIABLE(nJoystick1B);
         LOCK_VARIABLE(nJoystick1X);
         LOCK_VARIABLE(nJoystick1Y);
         LOCK_VARIABLE(nJoystick1CenterX);
         LOCK_VARIABLE(nJoystick1CenterY);
         LOCK_VARIABLE(nJoystick1Event1);
         LOCK_VARIABLE(nJoystick1Event2);
         LOCK_FUNCTION(mReadJoystick1Buttons);
         LOCK_FUNCTION(mReadJoystick1Axis);

         //install joystick 1 buttons
         nJoystick1Event1 = mCreateEvent((void (*)())&mReadJoystick1Buttons, 
                                         HzToMs(READ_BUTTONS_HZ));
         //could install?
         if (nJoystick1Event1 < 0)
            return(FALSE);

         //install joystick 1 axis
         nJoystick1Event2 = mCreateEvent((void (*)())&mReadJoystick1Axis, 
                                         HzToMs(READ_AXIS_HZ));
         //could install?
         if (nJoystick1Event2 < 0) {

            //uninstall and exit
            mDestroyEvent(nJoystick1Event1);
            return(FALSE);
         }

         //get callibration values for joystick 1
         mCallibrateJoystick(1);

      break;

      //install timed events for joystick 2:
      case 2:

         //lock my data and code
         LOCK_VARIABLE(nJoystick2Connected);
         LOCK_VARIABLE(nJoystick2Timeouts);
         LOCK_VARIABLE(nJoystick2A);
         LOCK_VARIABLE(nJoystick2B);
         LOCK_VARIABLE(nJoystick2X);
         LOCK_VARIABLE(nJoystick2Y);
         LOCK_VARIABLE(nJoystick2CenterX);
         LOCK_VARIABLE(nJoystick2CenterY);
         LOCK_VARIABLE(nJoystick2Event1);
         LOCK_VARIABLE(nJoystick2Event2);
         LOCK_FUNCTION(mReadJoystick2Buttons);
         LOCK_FUNCTION(mReadJoystick2Axis);

         //install joystick 2 buttons
         nJoystick2Event1 = mCreateEvent((void (*)())&mReadJoystick2Buttons, 
                                         HzToMs(READ_BUTTONS_HZ));
         //could install?
         if (nJoystick2Event1 < 0)
            //exit
            return(FALSE);

         //install joystick 2 axis
         nJoystick2Event2 = mCreateEvent((void (*)())&mReadJoystick2Axis, 
                                         HzToMs(READ_AXIS_HZ));
         //could install?
         if (nJoystick2Event2 < 0) {

            //uninstall and exit
            mDestroyEvent(nJoystick2Event1);
            return(FALSE);
         }

         //get callibration values for joystick 2
         mCallibrateJoystick(2);

      break;
   }

   //success!!!
   return(TRUE);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mUninstallJoysticks
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : uninstall all joysticks
// IN       : (none)
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mUninstallJoysticks() {

   //uninstall all events
   mDestroyEvent(nJoystick1Event1);
   mDestroyEvent(nJoystick1Event2);
   mDestroyEvent(nJoystick2Event1);
   mDestroyEvent(nJoystick2Event2);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mReadJoystick
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : read values from the given joystick number
// IN       : nJoystick = joystick number
//          : pnUp       = pointer to the up flag
//          : pnDown     = pointer to the down flag
//          : pnLeft     = pointer to the left flag
//          : pnRight    = pointer to the right flag
//          : pnA        = pointer to the button A flag
//          : pnB        = pointer to the button B flag
// OUT      : TRUE if the joystick is connected, FALSE if not
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
int mReadJoystick(int nJoystick, int *pnUp, int *pnDown, 
                  int *pnLeft, int *pnRight, int *pnA, int *pnB) {

   //valid joystick number?
   if ((nJoystick != 1)&&(nJoystick != 2)) {

      //invalid, reset values and exit!
      *pnUp = *pnDown = *pnLeft = *pnRight = *pnA = *pnB = FALSE;
      return(FALSE);
   }

   //joystick 1
   if (nJoystick == 1) {

      //connected?
      if (nJoystick1Connected == FALSE) {

         //not connected, reset values and exit!
         *pnUp = *pnDown = *pnLeft = *pnRight = *pnA = *pnB = FALSE;
         return(FALSE);
      }

      //get buttons
      *pnA = nJoystick1A;
      *pnB = nJoystick1B;

      //get directions
      *pnUp    = (nJoystick1Y < nJoystick1CenterY * AXIS_TOLERANCE_NEGATIVE) ?
                 TRUE : FALSE;
      *pnDown  = (nJoystick1Y > nJoystick1CenterY * AXIS_TOLERANCE_POSITIVE) ?
                 TRUE : FALSE;
      *pnLeft  = (nJoystick1X < nJoystick1CenterX * AXIS_TOLERANCE_NEGATIVE) ?
                 TRUE : FALSE;
      *pnRight = (nJoystick1X > nJoystick1CenterX * AXIS_TOLERANCE_POSITIVE) ?
                 TRUE : FALSE;
   }
   //joystick 2
   else  {

      //connected?
      if (nJoystick2Connected == FALSE) {

         //not connected, reset values and exit!
         *pnUp = *pnDown = *pnLeft = *pnRight = *pnA = *pnB = FALSE;
         return(FALSE);
      }

      //get buttons
      *pnA = nJoystick2A;
      *pnB = nJoystick2B;

      //get directions
      *pnUp    = (nJoystick2Y < nJoystick2CenterY * AXIS_TOLERANCE_NEGATIVE) ?
                 TRUE : FALSE;
      *pnDown  = (nJoystick2Y > nJoystick2CenterY * AXIS_TOLERANCE_POSITIVE) ?
                 TRUE : FALSE;
      *pnLeft  = (nJoystick2X < nJoystick2CenterX * AXIS_TOLERANCE_NEGATIVE) ?
                 TRUE : FALSE;
      *pnRight = (nJoystick2X > nJoystick2CenterX * AXIS_TOLERANCE_POSITIVE) ?
                 TRUE : FALSE;
   }

   //connected!
   return(TRUE);
}


/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
                         Joytick Register Information:
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                       Ŀ
                        7  6  5  4  3  2  1  0 
                       
                                              
 Joystick 2, Button B                       Joystick 1, X Axis
 Joystick 2, Button A                 Joystick 1, Y Axis
 Joystick 1, Button B           Joystick 2, X Axis
 Joystick 1, Button A     Joystick 2, Y Axis

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
