
/*


    /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
   /                                                                      \
   \      ___   _____   _____   _   __           _____   _____  __  __    /
   /     |_  | |___  | |  _  | | | / /          |  _  \ |  _  | \ \/ /    \
   \       | |  ___| | | | |_| | |/ /           | |_/ / | | | |  \  /     /
   /    _  | | |  _  | | |  _  |   <    in the  |  _ <  | | | |   ><      \
   \   | |_| | | |_| | | |_| | | |\ \           | |_| \ | |_| |  /  \     /
   /   |_____| |_____| |_____| |_| \_\          |_____/ |_____| /_/\_\    \
   \                                                                      /
   /                                                                      \
   \                                                                      /
   /             C    O    M    P    U    T    I    N    G                \
   \                                                                      /
    \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

*/


#define EXTERN
#include "game.h"

//============================================================================
// Module  : TEST
// Contents: testing timer and joystick functions
// Date    : 07.09.1997 (all dates in the "dd.mm.yyyy" format)
// Updates : (none)
//============================================================================

//event counters
volatile nEvent1Counter = 0;
volatile nEvent2Counter = 0;

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// mEventOne = just counts!
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mEventOne() { 

   //just increase counter!
   nEvent1Counter++; 
}
END_OF_FUNCTION(mEventOne); //used to lock this previous function!


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// mEventTwo = just counts!
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mEventTwo() { 

   //just increase counter!
   nEvent2Counter++; 
}
END_OF_FUNCTION(mEventTwo); //used to lock this previous function!


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// main = tests everything!
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void main() {

   //clear screen and put title
   clrscr();
   printf("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
   printf("  Testing \"Timer 1.0\" and \"Joystick 1.0\" - By JBC'1997\n");
   printf("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n\n");
   printf("-* move your joystick and press [Esc] to exit *- \n");

   //lock data and code
   LOCK_VARIABLE(nEvent1Counter);
   LOCK_VARIABLE(nEvent2Counter);
   LOCK_FUNCTION(mEventOne);
   LOCK_FUNCTION(mEventTwo);

   //startup timer
   mInstallTimer();

   //startup joysticks
   mInstallJoystick(1);
   mInstallJoystick(2);

   //create test events getting its handlers (time in milliseconds)
   int nEvent1 = mCreateEvent((void (*)())&mEventOne, 100); //at 10Hz
   int nEvent2 = mCreateEvent((void (*)())&mEventTwo, 200); //at 5Hz
   
   //get initial time now!
   mTimerStart();

   //main loop
   int nExit = FALSE;
   do {

      //print event values
      gotoxy(1,7);

      //print event 1 counter
      if (nEvent1 != -1)
         printf("RUNNING: ");
      else
         printf("STOPPED: ");
      printf("Event#1 at %02dhz = counting(%03d)\n", MsToHz(100),
             nEvent1Counter);

      //print event 2 counter
      if (nEvent2 != -1)
         printf("RUNNING: ");
      else
         printf("STOPPED: ");
      printf("Event#2 at %02dhz = counting(%03d)\n", MsToHz(200),
             nEvent2Counter);

      //get joystick 1 values
      int nConnection, nUp, nDown, nLeft, nRight, nButtonA, nButtonB;
      nConnection = mReadJoystick(1, &nUp, &nDown, &nLeft, &nRight, 
                                  &nButtonA, &nButtonB);

      //print joystick 1 values
      if (nConnection == TRUE)
         printf("\nJoystick 1 ON  = up%d down%d left%d right%d A%d B%d\n",
                nUp, nDown, nLeft, nRight, nButtonA, nButtonB);
      else
         printf("\nJoystick 1 OFF = up%d down%d left%d right%d A%d B%d\n",
                nUp, nDown, nLeft, nRight, nButtonA, nButtonB);

      //get joystick 2 values
      nConnection = mReadJoystick(2, &nUp, &nDown, &nLeft, &nRight, 
                                  &nButtonA, &nButtonB);
      
      //print joystick 2 values
      if (nConnection == TRUE)
         printf("Joystick 2 ON  = up%d down%d left%d right%d A%d B%d\n",
                nUp, nDown, nLeft, nRight, nButtonA, nButtonB);
      else
         printf("\nJoystick 2 OFF = up%d down%d left%d right%d A%d B%d\n",
                nUp, nDown, nLeft, nRight, nButtonA, nButtonB);

      //destroy "event 2" after 5 seconds (elapsed time since mTimerStart)
      if (mTimerStop() >= SToMs(5))
         nEvent2 = mDestroyEvent(nEvent2);

      //exit key?
      if (kbhit())
         if (getch() == 27)
            nExit = TRUE;

   //wait for the exit key
   } while(nExit == FALSE);

   //general program time (look for convertion tools in "timer.h")
   printf("\n...during %ds!\n", MsToS(mTimerStop()));

   //shutdown joystick
   mUninstallJoysticks();

   //shutdown timer
   mUninstallTimer();
}
