/*


    /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
   /                                                                      \
   \      ___   _____   _____   _   __           _____   _____  __  __    /
   /     |_  | |___  | |  _  | | | / /          |  _  \ |  _  | \ \/ /    \
   \       | |  ___| | | | |_| | |/ /           | |_/ / | | | |  \  /     /
   /    _  | | |  _  | | |  _  |   <    in the  |  _ <  | | | |   ><      \
   \   | |_| | | |_| | | |_| | | |\ \           | |_| \ | |_| |  /  \     /
   /   |_____| |_____| |_____| |_| \_\          |_____/ |_____| /_/\_\    \
   \                                                                      /
   /                                                                      \
   \                                                                      /
   /             C    O    M    P    U    T    I    N    G                \
   \                                                                      /
    \/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

*/


#define EXTERN extern
#include "game.h"

//============================================================================
// Module  : TIMER
// Contents: timer and event controlling
// Date    : 06.09.1997
// Updates : (none)
//============================================================================

//"private" methods
void  mTimerHandler  ();
char *mGetCMOSTime   ();

//PC timer definitions
#define IRQ0     0x8
#define PIT0     0x40
#define PITMODE  0x43
#define PITCONST 1193180L
#define PIT0DEF  18.2067597

//tick controlling
static float fTickPerMs      = 0.0182068;
static float fMsPerTick      = 54.9246551;
static float fFreq8h         = 18.2067597;
static BYTE  bTimerInstalled = FALSE;

//tick counters
volatile DWORD wwTicks8h;
volatile DWORD wwGetStart;

//semaphore to avoid reentrant interruptions
volatile BYTE bSemaphore = OFF;

//handler controlling
static _go32_dpmi_seginfo   stOldHandler, stNewHandler;
static _go32_dpmi_registers stAuxHandler;

//timer events
struct tagEvent {

   //function call
   void (*pfnEvent)();

   //milliseconds between each call, 
   //where -1 represents an unnused entry
   int nMilliseconds;

   //counter
   int nMsCounter;
};

//@CUSTOMIZE@ max number of events
#define MAX_TIMER_EVENTS 8

//events
volatile tagEvent astEvent[MAX_TIMER_EVENTS];
volatile int      nNumberOfEvents = 0;


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mTimerHandler
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : timer new handler and event caller
// IN       : (none)
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mTimerHandler(void) {

   //disable interruptions
   DISABLE();

   //reentrant interruption?
   if (bSemaphore == ON) {

      //ack the interruption and exit
      outportb(0x20, 0x20);
      return;
   }

   //turn on semaphore
   bSemaphore = ON;

   //count ticks
   wwTicks8h++;

   //event caller
   int nCurrent = 0;
   int nEvents  = nNumberOfEvents;
   while(nEvents > 0) {

      //activated entry?
      if (astEvent[nCurrent].nMilliseconds >= 0) {

         //countdown activated events
         nEvents--;

         //countdown timer
         astEvent[nCurrent].nMsCounter--;

         //right time?
         if (astEvent[nCurrent].nMsCounter <= 0) {

            //call event
            astEvent[nCurrent].pfnEvent();

            //restart counter
            astEvent[nCurrent].nMsCounter = astEvent[nCurrent].nMilliseconds;
         }
      }

      //next entry
      nCurrent++;
   }

   //turn off semaphore
   bSemaphore = OFF;

   //ack the interruption
   outportb(0x20, 0x20);
}
END_OF_FUNCTION(mTimerHandler);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mInstallTimer
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : install timer handler at 1000Hz
// IN       : (none)
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mInstallTimer() {

   //already installed?
   if (bTimerInstalled != TRUE) {

      //disable interruptions
      DISABLE();

      //lock data and code
      LOCK_VARIABLE(wwTicks8h);
      LOCK_VARIABLE(stAuxHandler);
      LOCK_VARIABLE(astEvent);
      LOCK_VARIABLE(nNumberOfEvents);
      LOCK_VARIABLE(bSemaphore);
      LOCK_FUNCTION(mTimerHandler);

      //install new handler
      _go32_dpmi_get_protected_mode_interrupt_vector(8, &stOldHandler);
      stNewHandler.pm_offset   = (int) mTimerHandler;
      stNewHandler.pm_selector = _go32_my_cs();
      _go32_dpmi_chain_protected_mode_interrupt_vector(8, &stNewHandler);

      //@CUSTOMIZE@ default Hz
      DWORD wwHz = 1000;

      //sometimes it doesn't seem to register if we only do this once!
      for(int nTries = 0; nTries < 4; nTries++) {

         //timer programming
         outportb(PITMODE, 0x36);
         DWORD wwPit0Value = PITCONST / wwHz;
         DWORD wwPit0Set   = (wwPit0Value & 0x00ff);
         outportb(PIT0, wwPit0Set);
         wwPit0Set = (wwPit0Value >> 8);
         outportb(PIT0, wwPit0Set);
      }

      //kickstart values
      bTimerInstalled = TRUE;
      fFreq8h         = wwHz;
      fTickPerMs      = fFreq8h / 1000;
      fMsPerTick      = 1000 / fFreq8h;
      wwTicks8h       = 0;
      wwGetStart      = 0;

      //reset events
      nNumberOfEvents = 0;
      for(int nEvent = 0; nEvent < MAX_TIMER_EVENTS; nEvent++)
         astEvent[nEvent].nMilliseconds = -1;

      //enable interruptions
      ENABLE();
   }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mUninstallTimer
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : deactivates timer handler
// IN       : (none)
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mUninstallTimer(void) {

   //already installed?
   if (bTimerInstalled == TRUE) {

      //disable interruptions
      DISABLE();

      //sometimes it doesn't seem to register if we only do this once!
      for(int nTries = 0; nTries < 4; nTries++) {

         //timer default programming
         outportb(PITMODE, 0x36);
         outportb(PIT0,    0x00);
         outportb(PIT0,    0x00);
       }

      //restore on handler
      _go32_dpmi_set_protected_mode_interrupt_vector(8, &stOldHandler);

      //enable interruptions
      ENABLE();

      //get CMOS time
      char *pszCMOSTime = mGetCMOSTime();
      DWORD wwTick = (DWORD)(PIT0DEF *
                             ((((float) *pszCMOSTime) * 3600) +
                              (((float) *(pszCMOSTime + 1)) * 60) +
                              (((float) *(pszCMOSTime + 2)))));

      //restore original timer with CMOS values
      biostime(1, wwTick);

      //default values
      bTimerInstalled = FALSE;
      fFreq8h         = PIT0DEF;
      fTickPerMs      = fFreq8h / 1000;
      fMsPerTick      = 1000 / fFreq8h;
   }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mTimerStart
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : startup the timer counter
// IN       : (none)
// OUT      : (none)
// Notes    : see mTimerStop
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mTimerStart() {

   //get starting tick
   wwGetStart = wwTicks8h;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mTimerStop
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : get elapsed time since last mTimerStart call
// IN       : (none)
// OUT      : milliseconds since last mTimerStart call
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
DWORD mTimerStop() {

   //get start and stop tick values
   DWORD wwStart = wwGetStart;
   DWORD wwStop  = wwTicks8h;

   //ok values?
   if (wwStop < wwStart)
      //no time...
      return 0;
   else {

      //get time
      DWORD wwDuration = wwStop - wwStart;
      DWORD wwMillisec = (DWORD)(wwDuration * fMsPerTick);
      return wwMillisec;
   }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mDelay
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : halt while milliseconds
// IN       : wwDelayMs = milliseconds to wait
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mDelay(DWORD wwDelayMs) {

   //startup values
   DWORD wwDelayBegin = 0;
   DWORD wwDelayEnd   = 0;
   DWORD wwDelayTick  = (DWORD)(wwDelayMs * fTickPerMs);

   //installed?
   if (bTimerInstalled == TRUE)
      wwDelayBegin = wwTicks8h;
   else
      wwDelayBegin = biostime(0, 0);

   //wait timer
   do {

      //installed?
      if (bTimerInstalled == TRUE)
         wwDelayEnd = wwTicks8h;
      else
         wwDelayEnd = biostime(0, 0);

   //until end of given time
   } while ((wwDelayEnd - wwDelayBegin) < wwDelayTick);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mGetCMOSTime
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : get CMOS time
// IN       : (none)
// OUT      : CMOS time in a buffer pointer
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
char *mGetCMOSTime(void) {

   //buffer and aux
   static char  szBuffer[6];
   char         chChar;

   //startup buffer and pointer
   char *pszBuff = szBuffer;
   memset(&stAuxHandler, 0, sizeof(stAuxHandler));
   stAuxHandler.h.ah = 0x02;
   _go32_dpmi_simulate_int(0x1a, &stAuxHandler);

   //get CMOS time
   chChar = stAuxHandler.h.ch;
   szBuffer[0] = (char)((int) (chChar & 0x0f) +
                        (int) ((chChar >> 4) & 0x0f) * 10);
   chChar = stAuxHandler.h.cl;
   szBuffer[1] = (char)((int) (chChar & 0x0f) +
                        (int) ((chChar >> 4) & 0x0f) * 10);
   chChar = stAuxHandler.h.dh;
   szBuffer[2] = (char)((int) (chChar & 0x0f) +
                        (int) ((chChar >> 4) & 0x0f) * 10);
   szBuffer[3] = stAuxHandler.h.dl;
   szBuffer[4] = (char)(stAuxHandler.x.flags & 0x0001);
   szBuffer[5] = 0x00;

   //return buffer pointer
   return(pszBuff);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mCreateEvent
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : startup a new timed event
// IN       : pfnAction     = address of the event function
//          : nMilliseconds = call this function each this milliseconds
// OUT      : handler number (see mDestroyEvent), or -1 on max number of
//          : handlers exceeded.
// Notes    : this event function must be locked (LOCK_FUNCTION) and all
//          : of its variables too (LOCK_VARIABLE). This function can't
//          : be "heavy" for the processor and can't call any OS function.
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
int mCreateEvent(void (*pfnAction)(), int nMilliseconds) {
   
   //timer installed?
   if (bTimerInstalled == FALSE)
      //not installed
      return(-1);

   //verify max events
   if (nNumberOfEvents >= MAX_TIMER_EVENTS)
      //limit exceeded!
      return(-1);

   //search for a free entry
   for(int nEntry = 0; nEntry < MAX_TIMER_EVENTS; nEntry++)
      //free?
      if (astEvent[nEntry].nMilliseconds < 0) {

         //put this event
         astEvent[nEntry].pfnEvent      = pfnAction;
         astEvent[nEntry].nMilliseconds = nMilliseconds;
         astEvent[nEntry].nMsCounter    = nMilliseconds;

         //increase number of events
         nNumberOfEvents++;

         //return success!!!
         return(nEntry);
      }

   //could not? error!
   return(-1);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mChangeEventTime
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : change a previously created event time
// IN       : nHandler      = handler of a previosly created event
//          : nMilliseconds = call this function each this milliseconds
// OUT      : (none)
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
void mChangeEventTime(int nHandler, int nMilliseconds) {
   
   //timer installed?
   if (bTimerInstalled == FALSE)
      //not installed
      return;

   //verify handler range
   if ((nHandler < 0)||(nHandler >= MAX_TIMER_EVENTS))
      //out of range!
      return;

   //valid entry?
   if (astEvent[nHandler].nMilliseconds < 0)
      //invalid entry!
      return;

   //new time is valid?
   if (nMilliseconds < 0)
      //invalid time value!
      return;

   //change entry time
   astEvent[nHandler].nMilliseconds =
      astEvent[nHandler].nMsCounter = nMilliseconds;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
// Function : mDestroyEvent
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Action   : deactivates a timed event
// IN       : nHandler = handler of a previously created event
// OUT      : -1, to use "nHandler = mDestroyEvent(nHandler);"
// Notes    : (none)
// Date     : 06.09.1997
// Updates  : (none)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
int mDestroyEvent(int nHandler) {

   //verify handler range
   if ((nHandler < 0)||(nHandler >= MAX_TIMER_EVENTS))
      //out of range!
      return(-1);

   //valid entry?
   if (astEvent[nHandler].nMilliseconds < 0)
      //invalid entry!
      return(-1);

   //free entry
   astEvent[nHandler].nMilliseconds = -1;

   //countdown event counter
   nNumberOfEvents--;

   //reset handler...
   return(-1);
}
